/* SPDX-License-Identifier: MIT
 * Copyright (c) 2020 Kingizor */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bd_internal.h"

void bd_set_error (const char*);



static int rb (struct COMPRESSOR *cmp) { /* Read byte */
    struct DATA_STREAM *in = &cmp->i;
    if (in->pos >= in->len) {
        cmp->flag |= READ_ERROR;
        return 0;
    }
    return in->stream[in->pos++];
}

static void wn (struct COMPRESSOR *cmp, int val) { /* Write nibble */
    struct DATA_STREAM *out = &cmp->o;
    if (out->pos >= out->len) {
        cmp->flag |= WRITE_ERROR;
        return;
    }
    if (out->half == 0) { /* Hi first */
        out->stream[out->pos] = val << 4;
    }
    else { /* Lo second */
        out->stream[out->pos++] |= val & 15;
    }
    out->half ^= 1;
}
static void wb (struct COMPRESSOR *cmp, int val) { /* Write byte */
    wn(cmp,  15 & (val >> 4));
    wn(cmp,  15 &  val);
}
static void ww (struct COMPRESSOR *cmp, int val) { /* Write word */
    wb(cmp, 255 & (val >> 8));
    wb(cmp, 255 &  val);
}




/* Certain cases have parameters ([a]ddr, [c]ount, etc) */
/* If a relevant case is active, choose_cases will set appropriate values */
struct CASE_WINDOW {
    unsigned short addr;
    unsigned char count;
};
struct CASE_PARAMS {
    struct CASE_WINDOW w[3]; /* 10, 11, 12 */
    unsigned short n;        /*  0 */
    unsigned char  rle;      /*  3,  4,  5 */
    unsigned char  wwin;     /*  9 */
    unsigned char  wlut;     /* 15 */
};

/* Determine which cases could be used for current block */
static unsigned choose_cases (struct COMPRESSOR *cmp, struct CASE_PARAMS *cp) {

    struct DATA_STREAM *in  = &cmp->i; /* Data to be compressed */
    struct DATA_STREAM *out = &cmp->o;
    unsigned char *is = &in->stream[in->pos];
    unsigned cases = 0;
    int i;

#define add_case(X) cases |= (1 << X)

    { /* RLE */
        int x  = in->len - in->pos;

        /* Check the next ~18 bytes */
        if (x > 18)
            x = 18;

        /* Are the next 3-18 bytes identical? */
        for (i = 1; i < x; i++)
            if (*is != is[i])
                break;

        /* If they are, then we can use an input byte for RLE */
        if (i > 2) {
            add_case(3);
            cp->rle = i;

            /* Or better yet, a constant */
            if (out->pos) {
                if (*is == out->stream[1])
                    add_case(4);
                if (*is == out->stream[2])
                    add_case(5);
            }
        }
    }

    { /* Byte Window */

        /* Window can be 8, 12 or 16 bits */
        cp->w[0].count = 0;
        cp->w[1].count = 0;
        cp->w[2].count = 0;

        for (i = 0; i < in->pos; i++) {

            int j = 18;
            int k;
            int pos = in->pos - i;

            /* How many bytes to compare */
            if (in->len-in->pos < j)
                j = in->len - in->pos;
            if (pos < j)
                j = pos;

            /* How many bytes match */
            for (k = 0; k < j; k++)
                if (is[k] != in->stream[i+k])
                    break;

            if (k > 2) {

                #define add_window(X,Y,Z) \
                    if (k > X.count) {\
                        X.count = k;\
                        X.addr  = pos - Z;\
                        add_case(Y);\
                    }

                     if (pos > 4095+259) {
                    add_window(cp->w[2], 12,   0); /* 16-bit */
                }
                else if (pos > 255+k) {
                    add_window(cp->w[1], 11, 259); /* 12-bit */
                }
                else {
                    add_window(cp->w[0], 10,   k); /*  8-bit */
                }
            }
            else if (k == 2 && pos < 18) { /* Word */
                cp->wwin = pos;
                add_case(9);
            }

            if (k)
                i += k-1;
        }
    }

    /* Byte Constants */
    if (out->pos) {
        if (*is == out->stream[3])
            add_case(7);
        if (*is == out->stream[4])
            add_case(8);
    }

    /* Word LUT */
    if (out->pos) {
        if (in->pos+1 < in->len) {
            if ((is[0] == out->stream[5])
            &&  (is[1] == out->stream[6]))
                add_case(6);

            for (i = 7; i < 38; i+=2) {
                if ((is[0] == out->stream[i])
                &&  (is[1] == out->stream[i+1])
                ) {
                    cp->wlut = (i - 7) / 2;
                    add_case(15);
                    break;
                }
            }
        }
    }

    { /* Repeats */

        /* Byte */
        if (in->pos && *is == is[-1])
            add_case(13);

        /* Word */
        if ((in->pos > 1)
        && ( in->pos + 1 < in->len)
        && (is[0] == is[-2])
        && (is[1] == is[-1]))
            add_case(14);

    }

    { /* Direct Copy */
        add_case(0); /* n  Bytes   */
        add_case(1); /* 1  Byte    */
        add_case(2); /* 1  Word    */
    }

    return cases;
}

/* How many bytes the current case outputs */
static int byte_count (struct CASE_PARAMS *cp, int c) {
    switch (c) {
        case  0:          { return cp->n;                  }
        case  1: case  7:
        case  8: case 13: { return 1;                      }
        case  2:
        case  6: case  9:
        case 14: case 15: { return 2;                      }
        case  3:
        case  4: case  5: { return cp->rle;                }
        case 10:
        case 11: case 12: { return cp->w[(c & 3)^2].count; }
    }
    return 0;
}

/* Ratios always include the command nibble */
static double ratio_case (struct CASE_PARAMS *cp, int c) {
    switch (c) {
        case  0:          { return (1.0 + cp->n) / cp->n;  }
        case  1:          { return 3.0 / 2.0;              }
        case  2:          { return 5.0 / 4.0;              }
        case  3:          { return 2.0 / cp->rle;          }
        case  4: case  5: { return 1.0 / cp->rle;          }
        case  6: case 14: { return 1.0 / 4.0;              }
        case  7:
        case  8: case 13:
        case  9: case 15: { return 1.0 / 2.0;              }
        case 10:          { return 2.0 / (cp->w[0].count); }
        case 11:          { return 2.5 / (cp->w[1].count); }
        case 12:          { return 3.0 / (cp->w[2].count); }
    }
    return 0;
}



static int encode_case (struct COMPRESSOR *cmp, struct CASE_PARAMS *cp, int c) {

    /* Write case */
    wn(cmp, c);

#define add_pos(X) cmp->i.pos += X

    switch (c) {

        /* n bytes */
        case 0: {
            wn(cmp, cp->n);
            while (cp->n--)
                wb(cmp, rb(cmp));
            break;
        }

        /* 2 bytes or 1 byte */
        case 2: { wb(cmp, rb(cmp)); }
        case 1: { wb(cmp, rb(cmp)); break; }

        /* RLE input (3-18) */
        case 3: {
            wn(cmp, cp->rle - 3);
            wb(cmp, rb(cmp));
            add_pos(cp->rle - 1);
            break;
        }

        /* RLE constant (3-18) */
        case 4:
        case 5: {
            wn(cmp, cp->rle - 3);
            add_pos(cp->rle);
            break;
        }

        /* Single constants or repeats */
        case 7: case  8: case 13: { add_pos(1); break; }
        case 6: case 14:          { add_pos(2); break; }

       /* Word window */
        case 9: {
            wn(cmp, cp->wwin - 2);
            add_pos(2);
            break;
        }

        /* 8-bit window */
        case 10: {
            wn(cmp, cp->w[0].count - 3);
            wb(cmp, cp->w[0].addr);
            add_pos(cp->w[0].count);
            break;
        }

        /* 12-bit window */
        case 11: {
            wn(cmp, cp->w[1].count -  3);
            wb(cmp, cp->w[1].addr  >> 4);
            wn(cmp, cp->w[1].addr  & 15);
            add_pos(cp->w[1].count);
            break;
        }

        /* 16-bit window */
        case 12: {
            wn(cmp, cp->w[2].count - 3);
            ww(cmp, cp->w[2].addr);
            add_pos(cp->w[2].count);
            break;
        }

        /* Word LUT */
        case 15: {
            wn(cmp, cp->wlut);
            add_pos(2);
            break;
        }
    }
    return 0;
}

/* Return the best non-copy case for the current position */
static int single_case (struct COMPRESSOR *cmp, struct CASE_PARAMS *cp) {
    unsigned short cases = choose_cases(cmp, cp);
    double best_ratio = 3;
    int best_case = 0; /* worst case by default */
    int i;

    /* Determine best case */
    for (i = 3; i < 16; i++) {
        double cur_ratio;
        
        if (!(cases & (1 << i)))
            continue;

        cur_ratio = ratio_case(cp, i);

        /* Pick the case with the best ratio */
        if (cur_ratio < best_ratio) {
            best_ratio = cur_ratio;
            best_case  = i;
        }

        /* If ratio is the same, pick the case that outputs most bytes */
        else if (cur_ratio == best_ratio) {
            if (byte_count(cp, i) > byte_count(cp, best_case)) {
                best_ratio = cur_ratio;
                best_case  = i;
            }
        }
    }
    return best_case;
}


/* Get the best case for the current position.
   If the best case is one of the copy cases (0/1/2),
   determine how many bytes we can skip.
   If we're evaluating data instead of processing it,
   we may want more data than case 0 can handle. (n > 15)
 */
static int get_case (struct COMPRESSOR *cmp, struct CASE_PARAMS *cp, int copy_limit) {

    struct DATA_STREAM *in = &cmp->i;
    int best_case = single_case(cmp, cp);

    /* If we have to copy, determine the best number */
    if (best_case < 3) {
        int i;

        /* Count how many consecutive bytes can't use a better case */
        for (i = 0; i < copy_limit; i++) {
            int new_case;
            if (in->pos >= in->len)
                break;
            new_case = single_case(cmp, cp);
            if ((new_case >= 3)
            &&  (ratio_case(cp, new_case) <= (3.0 / 4.0)))
                break;
            in->pos++;
        }
        in->pos -= i; /* Undo the lookahead */
        cp->n = i;    /* How many bytes to copy */
        best_case = (i > 2) ? 0 : i;
    }

    return best_case;
}



/* A compressed data block starts with a table of constants. */
/* First two entries are bytes used with RLE. */
/* Next three are two bytes and a word entry that have dedicated cases. */
/* Finally there are a further sixteen indexed words. */
struct DATA_CONSTANT {
    unsigned count;
    unsigned index;
};
static int sort_count (const void *aa, const void *bb) {
    const struct DATA_CONSTANT *b = aa;
    const struct DATA_CONSTANT *a = bb;
    if (a->count < b->count)
        return -1;
    if (a->count > b->count)
        return  1;

/* Here we face a dilemma. If the counts are the same we end up with
   multiple candidate constants, and there is no telling which ones will
   be the most effective without trying them. One option would be to
   decide randomly, but it's better to be deterministic. Better yet we
   could try and figure out what makes a particular choice better. */

    if (a->index < b->index)
        return  1;
    /* if (a->index > b->index) */
        return -1;
}
static void write_constant (
    struct COMPRESSOR *cmp,
    struct DATA_CONSTANT *dc,
    int dc_count,
    void (*write)(struct COMPRESSOR*, int),
    int write_count
) {
    int i;
    int units = 0;

    /* Sort by most frequent occurrences */
    qsort(dc, dc_count, sizeof(struct DATA_CONSTANT), sort_count);

    /* Write the n most frequently used units */
    for (i = 0; i < dc_count && units < write_count; i++, units++)
        write(cmp, dc[i].index);
}

static int choose_constants (struct COMPRESSOR *cmp) {

    struct DATA_STREAM *in = &cmp->i;
    struct DATA_CONSTANT *cb;
    struct DATA_CONSTANT *ib;
    struct DATA_CONSTANT *iw = calloc(256+256+65536, sizeof(struct DATA_CONSTANT));

    if (iw == NULL) {
        bd_set_error("Failed to allocate memory for constant search table.");
        return 1;
    }
    ib = iw + 65536;
    cb = ib +   256;

    {
        int i;
        for (i = 0; i < 256; i++)
            cb[i].index = i;
        for (i = 0; i < 256; i++)
            ib[i].index = i;
        for (i = 0; i < 65536; i++)
            iw[i].index = i;
    }

    while (in->pos < in->len) {

        /* Determine the best non-constant case */
        struct CASE_PARAMS cp;
        unsigned best_case = get_case(cmp, &cp, in->len);
        unsigned char *is = &in->stream[in->pos];
        int i;

        switch (best_case) {
            case 0: {
                for (i = 0; i < cp.n; i++)
                    ib[is[i]].count++;
                for (i = 0; i < cp.n/2; i++)
                    iw[(is[i*2] << 8) | is[i*2+1]].count++;
                break;
            }
            case 2: {
                iw[(*is << 8) | is[1]].count++;
                ib[ is[1]].count++;
            }
            case 1: { ib[*is].count++; break; }
            case 3: { cb[*is].count++; break; }
        }
        in->pos += byte_count(&cp, best_case);
    }

    /* Reset input position */
    in->pos = 0;

    /* Data starts with an unused value */
    /* Note: This might be a mistake on our end.
             If it is, delete this line */
    wb(cmp, 0);

    /* Write constants to output buffer */
    write_constant(cmp, cb,   256, wb,  2); /* Best   contiguous bytes */
    write_constant(cmp, ib,   256, wb,  2); /* Best incontiguous bytes */
    write_constant(cmp, iw, 65536, ww, 17); /* Best incontiguous words */

    free(iw);
    return 0;
}


static int error_check (int e) {
    if (e & READ_ERROR) {
        bd_set_error("Error reading from input stream.");
        return 1;
    }
    if (e & WRITE_ERROR) {
        bd_set_error("Output size has grown larger than the input size.");
        return 1;
    }
    return 0;
}

static int simple_method (struct COMPRESSOR *cmp) {
    for (;;) {
        struct CASE_PARAMS cp;

        if (encode_case(cmp, &cp, get_case(cmp, &cp, 15)))
            return 1;
        if (error_check(cmp->flag))
            return 1;
        if (cmp->i.pos >= cmp->i.len)
            return 0;
    }
}

int bd_compress (struct COMPRESSOR *cmp) {

    /* Generate the LUT */
    if (choose_constants(cmp))
        return 1;

    if (simple_method(cmp))
        return 1;

    /* End with a nul command */
    wb(cmp, 0);
    if (cmp->o.half)
        wn(cmp, 0);

    return 0;
}

