/* SPDX-License-Identifier: MIT
 * Copyright (c) 2020 Kingizor */

#include <stdio.h>
#include <stdlib.h>

#include "bd_internal.h"


/* Error reporting functions */
static const char *bd_error_msg = "";

const char *bd_get_error (void) {
    return bd_error_msg;
}
void bd_set_error (const char *error) {
    bd_error_msg = error;
}



/* File/Buffer handling */
static int check_size_comp (size_t input_size) {
    if (input_size > 0x10000) {
        bd_set_error("Input is too large to fit in a 16-bit bank. (n > 65536)");
        return 1;
    }
    if (input_size < 128) {
        bd_set_error("Input is too small. (n < 128)");
        return 1;
    }
    return 0;
}
static int check_size_decomp (size_t input_size) {
    if (input_size > 16777216) {
        bd_set_error("Input file is too large.");
        return 1;
    }
    if (input_size < 0x28) {
        bd_set_error("Input is too small. (n < 0x28)");
        return 1;
    }
    return 0;
}
static int check_input_mem (unsigned char *input, size_t input_size, int mode_comp) {
    if (input == NULL) {
        bd_set_error("Input pointer is NULL.");
        return 1;
    }
    if (mode_comp) {
        if (check_size_comp(input_size))
            return 1;
    }
    else {
        if (check_size_decomp(input_size))
            return 1;
    }
    return 0;
}

static int open_input_file (const char *fn, unsigned char **input, int *input_size, int ofs, int mode_comp) {
    FILE *f;
    long input_length;

    f = fopen(fn, "rb");
    if (f == NULL) {
        bd_set_error("Failed to open input file.");
        return 1;
    }

    if ((fseek(f, 0,   SEEK_END)  == -1)
    || ((input_length = ftell(f)) == -1)
    ||  (fseek(f, ofs, SEEK_SET)  == -1)) {
        bd_set_error("Failed to seek the input file.");
        goto error;
    }

    if (input_length <= ofs) {
        bd_set_error("File ofs is larger than the input file.");
        goto error;
    }

    if (mode_comp) {
        if (check_size_comp(input_length))
            goto error;
        *input_size = input_length;
    }
    else {
        if (check_size_decomp(input_length))
            goto error;
        if (ofs > input_length) {
            bd_set_error("Supplied decompression offset is larger than the input size.");
            goto error;
        }
        if (ofs < 0) {
            bd_set_error("Supplied decompresison offset is negative.");
            goto error;
        }
        *input_size = input_length - ofs;
        if (*input_size > 0x20028) {
            *input_size = 0x20028;
        }
        if (*input_size < 0x28) {
            bd_set_error("File size minus offset is too small.");
        }
    }

    *input = malloc(*input_size);
    if (*input == NULL) {
        bd_set_error("Failed to allocate memory for input buffer.");
        goto error;
    }

    if (fread(*input, 1, *input_size, f) != (size_t)*input_size) {
        bd_set_error("Failed to read from input file.");
        free(*input);
        goto error;
    }
    return 0;
error:
    if (f != NULL)
        fclose(f);
    return 1;
}

static int write_output_file (const char *fn, unsigned char *output, size_t output_size) {

    FILE *f;

    f = fopen(fn, "wb");
    if (f == NULL) {
        bd_set_error("Failed to open file for writing.");
        return 1;
    }

    if (fwrite(output, 1, output_size, f) != output_size) {
        bd_set_error("An error occurred while writing the output file.");
        fclose(f);
        return 1;
    }

    fclose(f);
    return 0;
}
static int open_output_buffer (unsigned char **output, size_t output_size) {
    *output = calloc(output_size, 1);
    if (*output == NULL) {
        bd_set_error("Failed to allocate memory for output buffer.");
        return 1;
    }
    return 0;
}



/* Compression handlers */

int bd_compress_mem_to_mem (
    unsigned char **output,
    size_t *output_size,
    unsigned char *input,
    size_t input_size
) {
    struct COMPRESSOR cmp = {};

    if (check_input_mem(input, input_size, 1))
        return 1;

    cmp.i.stream = input;
    cmp.i.len    = input_size;
    cmp.o.len    = cmp.i.len * 2 + 0x28;

    if (open_output_buffer(&cmp.o.stream, cmp.o.len))
        return 1;
    if (bd_compress(&cmp)) {
        free(cmp.o.stream);
        return 1;
    }
    *output      = cmp.o.stream;
    *output_size = cmp.o.pos;
    return 0;
}

int bd_compress_file_to_mem (
    unsigned char **output,
    size_t *output_size,
    const char *file_in
) {
    struct COMPRESSOR cmp = {};

    if (open_input_file(file_in, &cmp.i.stream, &cmp.i.len, 0, 1))
        goto error;

    cmp.o.len = cmp.i.len * 2 + 0x28;

    if (open_output_buffer(&cmp.o.stream, cmp.o.len))
        goto error;

    if (bd_compress(&cmp))
        goto error;

    free(cmp.i.stream);
    *output      = cmp.o.stream;
    *output_size = cmp.o.pos;
    return 0;
error:
    free(cmp.i.stream);
    free(cmp.o.stream);
    return 1;
}


int bd_compress_mem_to_file (
    const char *file_out,
    unsigned char *input,
    size_t input_size
) {
    unsigned char *output;
    size_t output_size;

    if (bd_compress_mem_to_mem(&output, &output_size, input, input_size))
        return 1;
    if (write_output_file(file_out, output, output_size)) {
        free(output);
        return 1;
    }
    free(output);
    return 0;
}

int bd_compress_file_to_file (
    const char *file_out,
    const char *file_in
) {
    unsigned char *output;
    size_t output_size;

    if (bd_compress_file_to_mem(&output, &output_size, file_in))
        return 1;
    if (write_output_file(file_out, output, output_size)) {
        free(output);
        return 1;
    }
    free(output);
    return 0;
}





/* Decompression handlers */

int bd_decompress_mem_to_mem (
    unsigned char **output,
    size_t *output_size,
    unsigned char *input,
    size_t input_size
) {
    struct DECOMPRESSOR dc = {};

    if (check_input_mem(input, input_size, 0))
        return 1;

    dc.input     = input;
    dc.input_len = input_size;

    if (open_output_buffer(&dc.output, 0x10000))
        return 1;

    if (bd_decompress(&dc)) {
        free(dc.output);
        return 1;
    }
    *output = dc.output;
    *output_size = dc.x;
    return 0;
}
int bd_decompress_file_to_mem (
    unsigned char **output,
    size_t *output_size,
    const char *file_in,
    int position
) {
    struct DECOMPRESSOR dc = {};

    if (open_input_file(file_in, &dc.input, &dc.input_len, position, 0))
        goto error;
    if (open_output_buffer(&dc.output, 0x10000))
        goto error;
    if (bd_decompress(&dc))
        goto error;

    free(dc.input);
    *output      = dc.output;
    *output_size = dc.x;
    return 0;
error:
    free(dc.input);
    free(dc.output);
    return 1;
}

int bd_decompress_mem_to_file (
    const char *file_out,
    unsigned char *input,
    size_t input_size
) {
    unsigned char *output;
    size_t output_size;

    if (bd_decompress_mem_to_mem(&output, &output_size, input, input_size))
        return 1;
    if (write_output_file(file_out, output, output_size)) {
        free(output);
        return 1;
    }
    free(output);
    return 0;
}
int bd_decompress_file_to_file (
    const char *file_out,
    const char *file_in,
    size_t position
) {
    unsigned char *output;
    size_t output_size;

    if (bd_decompress_file_to_mem(&output, &output_size, file_in, position))
        return 1;
    if (write_output_file(file_out, output, output_size)) {
        free(output);
        return 1;
    }
    free(output);
    return 0;
}


