/* SPDX-License-Identifier: MIT
 * Copyright (c) 2020 Kingizor */ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bd_internal.h"

static void write_out (struct DECOMPRESSOR *dc, unsigned char val) {
    if (dc->x > 0xFFFF) {
        bd_set_error("Attempted to write out of bounds.");
        dc->flag |= WRITE_ERROR;
        return;
    }
    dc->output[dc->x++] = val;
}
static unsigned char read_out (struct DECOMPRESSOR *dc, unsigned short v) {

    int addr = dc->x - v;

    if (addr < 0 || addr > 0xFFFF) {
        bd_set_error("Attempted to read out of bounds.");
        dc->flag |= READ_ERROR;
        return 0;
    }

    return dc->output[addr];
}

static int rn (struct DECOMPRESSOR *dc) { /* read nibble */
    if (dc->y >= dc->input_len) {
        bd_set_error("Attempted to read past end of input.");
        dc->flag |= READ_ERROR;
        return 0;
    }

    dc->half ^= 1;

    if (dc->half)
        return dc->input[dc->y  ] >> 4; /* hi 1st */
        return dc->input[dc->y++] & 15; /* lo 2nd */
}
static int rb (struct DECOMPRESSOR *dc) { /* read byte */
    int hi = rn(dc) << 4;
    int lo = rn(dc);
    return hi|lo;
}

static int decode_case (struct DECOMPRESSOR *dc) {

    int c = rn(dc);

    switch (c) {

        /* Copy n bytes */
        case 0: {
            int i = rn(dc);

            if (!i) {
                dc->flag |= DECOMP_QUIT;
                break;
            }

            while (i--) 
                write_out(dc, rb(dc));

            break;
        }

        /* Write two bytes */
        case 2: {
            write_out(dc, rb(dc));
        }

        /* Write a byte */
        case 1: {
            write_out(dc, rb(dc));
            break;
        }

        /* Write a byte 3-18 */
        case 3: {
            int i = rn(dc) + 3;
            int z = rb(dc);
            while (i--)
                write_out(dc, z);
            break;
        }

        /* Write a constant 3-18 */
        case  4: case  5: {
            int i = rn(dc) + 3;
            while (i--)
                write_out(dc, dc->input[1 + (c & 1)]);
            break;
        }

        /* Write a word constant */
        case 6: {
            write_out(dc, dc->input[5]);
            write_out(dc, dc->input[6]);
            break;
        }

        /* Write a byte constant */
        case 7: case 8: {
            write_out(dc, dc->input[3 + ((c ^ 1) & 1)]);
            break;
        }

        /* Write a recent word */
        case 9: {
            int addr = rn(dc) + 2;
            write_out(dc, read_out(dc, addr));
            write_out(dc, read_out(dc, addr));
            break;
        }

        /* 8-bit window */
        case 10: {
            int i    = rn(dc) + 3;
            int addr = rb(dc) + i;
            while (i--)
                write_out(dc, read_out(dc, addr));
            break;
        }

        /* 12-bit window */
        case 11: {
            int i     = rn(dc) + 3;
            int addr  = rb(dc) << 4;
                addr |= rn(dc);
                addr += 0x103;
            while (i--)
                write_out(dc, read_out(dc, addr));
            break;
        }

        /* 16-bit window */
        case 12: {
            int i     = rn(dc) + 3;
            int addr  = rb(dc) << 8;
                addr |= rb(dc);
            while (i--)
                write_out(dc, read_out(dc, addr));
            break;
        }

        /* Repeat last byte */
        case 13: {
            write_out(dc, read_out(dc, 1));
            break;
        }

        /* Repeat last word */
        case 14: {
            write_out(dc, read_out(dc, 2));
            write_out(dc, read_out(dc, 2));
            break;
        }

        /* Word LUT */
        case 15: {
            int addr = (rn(dc) << 1) + 7;
            write_out(dc, dc->input[addr++]);
            write_out(dc, dc->input[addr  ]);
            break;
        }
    }
    return 0;
}

int bd_decompress (struct DECOMPRESSOR *dc) {

    dc->x  = 0;
    dc->y  = 0x27;

    for (;;) {
        if (decode_case(dc))
            return 1;
        if (dc->flag & READ_ERROR)
            return 1;
        if (dc->flag & DECOMP_QUIT)
            return 0;
    }
}

